#lang scheme

;Created by: Alex Laird
;Purpose: Project #2 in Survey of Programming Languages class
;Date Created: 2/4/09
;Filename: adlibmcp.ss
;Version: 0.1
;
;The following is an application that emulates an IBM 650, one of the
;few truly base-10 based hardware computers ever made.
;
;The computer would read programs as commands in the form of
;~# ### ### ### where ~ was a + or - and the number blocks were memory
;locations.  The first ~# was the operation to be performed on the
;given numbers.  The second ### was the memory location of the first
;number, the third ### was the memory location of the second number,
;and the fourth ### was the destination memory location, where the
;answer would be stored usually.
;
;The commands are as follows:
;
;f = function
;s = sign
;
;f\s |       +       |         -
;--------------------|------------
; 0  |     move      | negate and move
; 1  |       +       |         -
; 2  |       x       |         /
; 3  |    square     |   square root
; 4  |      =        |        !=
; 5  |      >=       |         <
; 6  |   x(y) -> z   |     x -> y(z)
; 7  | incr and test |     (unused)
; 8  |     read      |       print
; 9  |     stop      |     (unused)
;
;The program is read in in three sections; initial data values (the memory)
;program instructions, and input data.  The first section contains
;initialization for certain memory locations.  The second section is where
;the messages are stored which are to manipulate the memory locations, and
;the third section, if used, is used on a +8 read operation.

;store the command-line arguments
(define argv (vector->list (current-command-line-arguments)))
;define the input port, whether it be from a file (if command-line arguments
;were specified) or from the shell
(define input (if (and (>= (length argv) 1)
                       (not (equal? (list-ref argv 0) "-")))
                  (open-input-file (list-ref argv 0))
                  (current-input-port)))
;define the output port, whether it be from a file (if command-line arguments
;were specified) or from the shell
(define output (if (and (>= (length argv) 2)
                        (not (equal? (list-ref argv 1) "-")))
                   (open-output-file (list-ref argv 1) #:exists 'replace)
                   (current-output-port)))

;a procedure that reads a section of a data and stores until the section is
;closed by a "+9999999999" or the eof is reached
;if reading from the shell and not from a file, the inputs (final) section is
;closed when a "done" is reached
(define (read-section data location)
  (define line (read-line input))
  (when (not (or (equal? line "done")
                 (eof-object? line)
                 (equal? (substring line 0 11) "+9999999999")))
    (vector-set! data location (substring line 0 11))
    (read-section data (+ 1 location)))
  data)

;a procedure that ensures all values in the memory vector are numerical
;and not string types
(define (strings-to-numbers data index)
  (when (not (= index (vector-length data)))
    (when (string? (vector-ref data index))
      (when (not (equal? (vector-ref data index) "invalid"))
        (vector-set! data index (string->number (vector-ref data index)))))
    (strings-to-numbers data (+ 1 index)))
  data)

;a vector that is the virtual memory for the emulator
(define memory (strings-to-numbers (read-section (make-vector 1000 "invalid") 0) 0))
;a vector that contains the list of al commands read from the input
(define messages (read-section (make-vector 1000 "invalid") 0))
;a vector that contains any input values for the emulator (if specified)
;from the input
(define inputs (read-section (make-vector 1000 "invalid") 0))

;the main procedure that parses each command from the messages vector
;this procedure is not entirely functional due to the use of the "let"
;command
(define (run index-pointer memory messages inputs input-pointer)
  ;store the prefix, location1, location2, and destination values in temporary
  ;variables for significantly easier access during the parsing
  ((let ((prefix (substring (vector-ref messages index-pointer) 0 2))
         (location1 (string->number (substring (vector-ref messages index-pointer) 2 5)))
         (location2 (string->number (substring (vector-ref messages index-pointer) 5 8)))
         (destination (string->number (substring (vector-ref messages index-pointer) 8 11))))
     (lambda()
       (cond
         ((equal? prefix "+0")
          ;move
          (vector-set! memory destination (vector-ref memory location1))
          (run (+ index-pointer 1) memory messages inputs input-pointer))
         ((equal? prefix "+1")
          ;add
          (vector-set! memory destination (+ (vector-ref memory location1) (vector-ref memory location2)))
          (run (+ index-pointer 1) memory messages inputs input-pointer))
         ((equal? prefix "+2")
          ;multiply
          (vector-set! memory destination (* (vector-ref memory location1) (vector-ref memory location2)))
          (run (+ index-pointer 1) memory messages inputs input-pointer))
         ((equal? prefix "+3")
          ;square
          (vector-set! memory destination (* (vector-ref memory location1) (vector-ref memory location1)))
          (run (+ index-pointer 1) memory messages inputs input-pointer))
         ((equal? prefix "+4")
          ;equal
          (run (if (= (vector-ref memory location1) (vector-ref memory location2))
                   destination
                   (+ 1 index-pointer)) memory messages inputs input-pointer))
         ((equal? prefix "+5")
          ;greater or equal
          (run (if (>= (vector-ref memory location1) (vector-ref memory location2))
                   destination
                   (+ 1 index-pointer)) memory messages inputs input-pointer))
         ((equal? prefix "+6")
          ;x(y) -> z
          (vector-set! memory destination (+ location1 (vector-ref memory location2)))
          (run (+ index-pointer 1) memory messages inputs input-pointer))
         ((equal? prefix "+7")
          ;increment and test
          (vector-set! memory location1 (+ 1 (vector-ref memory location1)))
          (run (if (not (= (vector-ref memory location1) (vector-ref memory location2)))
                   destination
                   (+ 1 index-pointer)) memory messages inputs input-pointer))
         
         ((equal? prefix "+8")
          ;read
          (vector-set! memory destination (string->number (vector-ref inputs input-pointer)))
          (run (+ index-pointer 1) memory messages inputs (+ input-pointer 1)))
         ;the end
         ((equal? prefix "+9")
          0)
         ((equal? prefix "-0")
          ;negate and move
          (vector-set! memory destination (- 0 (vector-ref memory location1)))
          (run (+ index-pointer 1) memory messages inputs input-pointer))
         ((equal? prefix "-1")
          ;subtract
          (vector-set! memory destination (- (vector-ref memory location1) (vector-ref memory location2)))
          (run (+ index-pointer 1) memory messages inputs input-pointer))
         ((equal? prefix "-2")
          ;divide
          (vector-set! memory destination (/ (vector-ref memory location1) (vector-ref memory location2)))
          (run (+ index-pointer 1) memory messages inputs input-pointer))
         ((equal? prefix "-3")
          ;square root
          (vector-set! memory destination (sqrt(vector-ref memory location1)))
          (run (+ index-pointer 1) memory messages inputs input-pointer))
         ((equal? prefix "-4")
          ;not equal
          (run (if (not (= (vector-ref memory location1) (vector-ref memory location2)))
                   destination
                   (+ 1 index-pointer)) memory messages inputs input-pointer))
         
         ((equal? prefix "-5")
          ;less than
          (run (if (< (vector-ref memory location1) (vector-ref memory location2))
                   destination
                   (+ 1 index-pointer)) memory messages inputs input-pointer))
         ((equal? prefix "-6")
          ;x -> y(z)
          (vector-set! memory (+ location2 (vector-ref memory destination)) (vector-ref memory location1))
          (run (+ index-pointer 1) memory messages inputs input-pointer))
         ((equal? prefix "-8")
          ;write
          (display (vector-ref memory location1) output)
          (newline output)
          (run (+ index-pointer 1) memory messages inputs input-pointer))
         (else 1))))))

;initial emulator call
(when (= (run 0 memory messages inputs 0) 1)
  (display "Somehow, someway, an invalid prefix was given to the emuator; hence, it failed."))

;close input and output ports
(close-input-port input)
(close-output-port output)